/* sender for internet datagram demo */
/* modification 9/20/93 with Louis Sanchez's bug fix */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>



help_msg()
{
	printf("format: phase2 -[dr] sec usec hostname portno\n");
	printf("          option: d set debug mode\n");
	printf("          option: r set retry number\n");
}

main(argc, argv)
int argc;
char *argv[];
{
	int sock; 
	struct sockaddr_in to;
	struct sockaddr_in from;
	struct hostent *hp, *gethostbyname();

	int n;
	char toBuf[1024];
	char fromBuf[1024];
	char buf[1024];
	char hostname[64];
	extern char *optarg;
	extern int   optind;
	int option_exist = 0;
	int c;
	int debug = 0;
	char *rcv_name;  /* receiver's socket name */
	int toPort_no;
	int length;
	int flag = 1;

	fd_set readfds;
	struct timeval timeout;
	int state = 1;
	int retry = 3;
	int r;
	int maxfdpl;
	int i;

	/* process command line args */
	while ((c=getopt(argc,argv,"dr:")) != EOF)
	{
		option_exist = 1;
		switch(c)
		{
		case 'd':
			debug = 1;
			if (debug)
				printf("turn on debug mode.\n");
			break;

		case 'r':
			if ((retry = atoi(optarg)) < 0) retry = 0;
			if (debug)
				printf("The retry number is set to %d\n", retry);
			break;
		case '?':
			help_msg();
			exit (1);
		default:
			fprintf (stderr, "unrecognized arg >%s<\n", optarg);
			help_msg();
			exit (1);
		}
	}

	timeout.tv_sec = atol(argv[optind++]);
	timeout.tv_usec = atol(argv[optind++]);
	printf("timeout is set to %d sec %d usec\n", 
		timeout.tv_sec, timeout.tv_usec);
	rcv_name = argv[optind++];
	if ((toPort_no = atoi(argv[optind])) < 1025) toPort_no = 1025;

	if (debug) printf("The receiver's host name is %s, port_no=%d\n", rcv_name, toPort_no);


	/* create socket from which to read */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("opening datagram socket");
		exit(1);
	}
	/* create name structure with wildcard using INADDR_ANY */
	from.sin_family = AF_INET;
	from.sin_addr.s_addr = INADDR_ANY;
	from.sin_port = 0;  
	/* port no. < 1024 IPPORT_RESERVED is for privillege process */
	/* here just initialized it, it will be assigned value after call */
	/* getsockname() */
	if (bind(sock, &from, sizeof(struct sockaddr_in))) {
		perror("binding name to datagram socket");
		exit(1);
	}
	length = sizeof(from);
	if (getsockname(sock, &from, &length)) {
		perror("getting socket name");
		exit(1);
	}
	/* Find assigned port value and print out. */
	printf("socket has port #%d\n", ntohs(from.sin_port));

	/* construct name of the socket to send to 
	 * gethostbyname() returns a structure including the network address
	 * of the specified host from /etc/hosts. The port number is taken from
	 * the cmd line
	 * create name structure for the UNIX domain socket 
	 * use the name from cmd argument 
	 */
	hp = gethostbyname(rcv_name);
	if (hp == 0) {
		fprintf(stderr, "%s: unknown host",rcv_name);
		exit(2);
	}
	bcopy(hp->h_addr, &to.sin_addr, hp->h_length);
	to.sin_family = AF_INET;
	to.sin_port = htons(toPort_no);

	while (flag) {
	printf("Enter state %d\n", state);
	switch(state) {
	case 1:
		printf("Enter the message=");
		gets(toBuf);
		if (strcmp(toBuf, "$") == 0) state = 4;
		else {
			state = 2;
			r = retry;
			}
		break;
	case 2:
		if (sendto(sock, toBuf, strlen(toBuf), 0, &to, sizeof(to)) < 0) 
			perror("sending datagram message");
		state = 3;
		break;
	case 3:
		FD_ZERO(&readfds);
		FD_SET(0, &readfds);
		FD_SET(sock, &readfds);
		maxfdpl = sock+1;
		if ((i=select(maxfdpl, &readfds, 0, 0, &timeout)) < 0)  {
			perror("select error"); exit(1);
		}
		if (i==0) {
			printf("timeout!");
			if (r == 0) state = 4;
			else {
				r--;
				printf("retry %d\n", r);
				state = 2;
			}
		}
		if (FD_ISSET(0, &readfds)) {
			gets(buf);
			if (strcmp(buf, "$") == 0) state = 4;
			else {
				printf("user input =%s != $, neglect it\n", buf);
			}
		}
		if (FD_ISSET(sock, &readfds)) {
			n = recv(sock, fromBuf, sizeof(fromBuf), 0);
			if (n <= 0) {
				perror("error in recv()"); exit(1);
			}
			printf("received ack =%s\n", fromBuf);
			state = 1;
		}
		break;
	case 4:
		flag = 0;
		printf("exit to while loop\n");
		break;
	default:
		printf("incorrect state!\n");
		flag = 0;
		break;
	}
	}
	close(sock);
}

